# -*- coding: utf-8 -*-
################################################################################
# pcollections/util/__init__.py
# Initialization file for the pcollections utilities.
# By Noah C. Benson

"""Utility functions for persistent and transient collections.
"""

from ._core import (setcmp, seqcmp, seqstr)
